type terme =
   V of string |
   C of string |
   U of string * terme |
   B of string * terme * terme;;

exception Erreur;;

type 'a algbre = { eval_c : string -> 'a;
                    eval_u : string -> 'a -> 'a;
                    eval_b : string -> 'a -> 'a -> 'a };;

let value {eval_c = evc; eval_u = evu; eval_b = evb} terme liaisons =
   let rec ev = function
      V x -> List.assoc x liaisons |
      C s -> evc s |
      U (s,u) -> evu s (ev u) |
      B (s,u,v) -> evb s (ev u) (ev v) in
   ev terme;;

type qualit =
   Constante | Variable | Prfixe | Postfixe | Gassoc | Dassoc | Nassoc;;

let imprime_inf priorit qualit =
   let pr = function
      V x -> priorit x - 1 |
      C f -> priorit f |
      U(f,_) -> priorit f |
      B(f,_,_) -> priorit f in
   let rec imp =
      let i parenthse t =
         if parenthse then print_char '(';
         imp t;
         if parenthse then print_char ')' in
   function
      V x -> print_string x |
      C f -> print_string f |
      U(f,u) -> (match qualit (priorit f) with
                   Prfixe -> print_string f; i (pr u < priorit f) u |
                   Postfixe -> i (pr u < priorit f) u; print_string f) |
      B(f,u,v) ->
         let q = qualit (priorit f) in
         i (pr u < priorit f or pr u = priorit f & q <> Gassoc) u;
         print_string f;
         i (pr v < priorit f or pr v = priorit f & q <> Dassoc) v in
   imp;;

let analyse_inf prioritmax priorit qualit l =
   let u = Array.of_list (l @ ["$"])
   and k = ref 0 in
   let rec expr p =
      if p = prioritmax
         then
            match u.(!k) with
               "(" -> incr k;
                      let t = expr 1 in
                      if u.(!k) <> ")" then raise Erreur;
                      incr k;
                      t |
               f -> incr k;
                    if priorit f = prioritmax then C f else V f
         else
            match qualit p with
               Prfixe -> let f = u.(!k) in
                          if priorit f = p
                             then (incr k; U(f,expr p))
                             else expr (p + 1) |
               Postfixe -> let t = ref (expr (p + 1)) in
                           while priorit u.(!k) = p do
                              t := U(u.(!k),!t);
                              incr k
                           done;
                           !t |
               Gassoc -> let t = ref (expr (p + 1)) in
                         while priorit u.(!k) = p do
                            let f = u.(!k) in
                            incr k;
                            t := B(f,!t,expr (p + 1))
                         done;
                         !t |
               Dassoc -> let t = expr (p + 1) in
                         if priorit u.(!k) = p
                            then let f = u.(!k) in incr k; B(f,t,expr p)
                            else t |
               Nassoc -> let t = expr (p + 1) in
                         if priorit u.(!k) = p
                            then let f = u.(!k) in incr k; B(f,t,expr(p+1))
                            else t in
   let t = expr 1 in if u.(!k) = "$" then t else raise Erreur;;

type er = terme;;

let  priorit_rationnelle = function
   "+"  -> 1 | "" -> 2 | "*" -> 3 | _ ->4
and qualit_rationnelle = function
   1 | 2 -> Gassoc | 3 -> Postfixe | 4 -> Constante;;

let (er_print : er -> unit) =
   imprime_inf priorit_rationnelle qualit_rationnelle;;

let er_of_string s =
   let n = String.length s in
   let lexmes =
      let rec lexic i fin =
         if i = n  then []
                   else let x = String.make 1 s.[i] in
                        let l = x :: lexic (i + 1) (x <> "+" & x <> "(") in
                        if fin & x <> "*" & x <> "+" & x <> ")"
                           then "" :: l else l in
      lexic 0 false in
   (analyse_inf 4 priorit_rationnelle qualit_rationnelle lexmes : er);;

type 'a t = ('a -> 'a -> int) * 'a list;; (* ensembles *)

let empty c = c,[]

and is_empty (_,l) = l = []

and mem x (comp,e) =
   let rec app = function
       [] -> false
     | a :: e -> let c = comp x a in c = 0 || (c > 0 && app e) in
   app e

and add x (comp,e) =
  let rec aj = function
      [] -> [x]
    | a :: e as f ->  
        let c = comp x a in
        if c < 0 then x :: f 
        else if c = 0 then f 
        else a :: aj e in
  (comp,aj e)
  
and remove x (comp,e) =
  let rec supp = function
      [] -> []
    | a :: e as f ->  
        let c = comp x a in
        if c < 0 then f
        else if c = 0 then e
        else a :: supp e in
  (comp,supp e)

and union (comp,e) (_,f) =
   let rec fusion = function
    | [],f -> f
    | e,[] -> e
    | (a :: e as e'),(b :: f as f') ->
        let c = comp a b in
        if c < 0 then a :: fusion (e,f')
        else if c = 0 then a :: fusion (e,f)
        else b :: fusion (e',f) in
  comp,fusion (e,f)
  
and inter (comp,e) (_,f) =
   let rec inters = function
      [], f -> []
    | e, [] -> []
    | (a :: e as e'), (b :: f as f') ->
         let c = comp a b in
         if c < 0 then inters (e, f')
         else if c = 0 then a :: inters (e, f)
         else inters (e', f) in
  comp,inters(e,f)
  
and diff (comp,e) (_,f) =
   let rec differ = function
      [], f -> []
    | e, [] -> e
    | (a :: e as e'), (b :: f as f') ->
         let c = comp a b in
         if c < 0 then a :: differ (e, f')
         else if c = 0 then differ (e, f)
         else differ (e', f) in
  comp,differ(e,f)

and compare (comp,e) (_,f) =
   let rec cpr = function
    | [], [] -> 0
    | [], f -> -1
    | e, [] -> 1
    | (a :: e), (b :: f) ->
         let c = comp a b in if c = 0 then cpr (e,f) else c in
   cpr(e,f)

and elements (_,e) = e

and iter f (_,e) = List.iter f e;;

let equal e f = compare e f = 0;;

type tat = {mutable transitions : (char * tat) list; numro : int}
and af = {initial : tat t; final : tat t; contient_mu : bool};;

let compteur_d'tats = ref 0;;

let vide = empty (fun {numro = n} {numro = m} -> n - m);;

let af_delta q x = (* tat t -> char -> tat t *)
      let q' = ref vide in
      iter (function {transitions = l} ->
        List.iter (function (y,e) ->
          if x = y then q' := add e !q') l) q;
      !q';;

let af_rec {initial = i; final = f} u =
   let q = ref i in
   for j = 0 to String.length u - 1 do
      q := af_delta !q u.[j]
   done;
   not (inter !q f = vide);;

let nouvel_tat() =
   incr compteur_d'tats;
   {transitions = []; numro = !compteur_d'tats}
and ajoute_transition e c e' = e.transitions <- (c,e') :: e.transitions;;

let af_vide = {initial = vide; final = vide; contient_mu = false}
and af_motvide() = let e = add (nouvel_tat()) vide in
                   {initial = e; final = e; contient_mu = true}
and af_car c =
   let e = nouvel_tat() and e' = nouvel_tat() in
   ajoute_transition e c e';
   {initial = add e vide; final = add e' vide; contient_mu = false};;

let af_add t t'=
        {initial     = union t.initial t'.initial;
         final       = union t.final t'.final;
         contient_mu = t.contient_mu or t'.contient_mu};;

let accole t t' =
   iter (function f ->
      iter (function i' ->
         List.iter (function (c,q) ->
            ajoute_transition f c q) i'.transitions) t'.initial) t.final;;

let af_mult t t' =
   accole t t';
   {initial = t.initial;
    final = if t'.contient_mu then union t.final t'.final else t'.final;
    contient_mu = t.contient_mu & t'.contient_mu };;

let af_toile t =
   let e = nouvel_tat() in
   accole t t;
   {initial = add e t.initial; final = add e t.final; contient_mu = true};;

let af_of_er (e : er) = value
   {eval_c = (function "V" -> af_vide | "" -> af_motvide() | c -> af_car c.[0]);
    eval_u = (function "*" -> af_toile | _ -> raise Erreur);
    eval_b = (function "+" -> af_add | _ -> af_mult)} e [];;


type alphabet =
    {mutable nombre_lettres : int;
     mutable lettres        : char array;
     mutable numero         : (char,int) Hashtbl.t};;

let alphabet_global =
   {nombre_lettres = 0; lettres = [||]; numero = Hashtbl.create 11};;

let dfinir_alphabet s =
  let n = String.length s in
  let t = Array.init n (function i -> s.[i]) in
  alphabet_global.nombre_lettres <- n;
  alphabet_global.lettres <- t;
  Hashtbl.clear alphabet_global.numero;
  for i = 0 to Array.length t-1 do
     Hashtbl.add alphabet_global.numero t.(i) i
  done;;

let nombre_de_lettres() = alphabet_global.nombre_lettres
and lettre_of_numro i = alphabet_global.lettres.(i)
and numro_of_lettre c = Hashtbl.find alphabet_global.numero c;;

type afd = {delta : int array array; accept : bool array};;

let afd_rec {delta = d; accept = a} u =
   let q = ref 0 in
   for j = 0 to String.length u - 1 do
      q := d.(!q).(numro_of_lettre u.[j])
   done;
   a.(!q);;

let construire_afd gal dpart delta final =
(* ('a -> 'a -> bool) -> 'a -> ('a -> char -> 'a) -> ('a -> bool) -> afd *)
   let t = ref []          (* liste (tat,numro).                          *)
   and f = Queue.create()    (* file des tats  traiter.                     *)
   and delta' = ref []     (* liste des vecteurs delta'.(0), delta'.(1),___ *)
   and accept = ref []     (* liste  accept(0),___                          *)
   and n = ref 0 in        (* nombre d'tats.                               *)
   let numro_of_tat q =
      (* 'a -> int. (l'tat q est rajout si ncessaire) *)
      let rec cherche = function
         [] -> raise Not_found |
         (q',p) :: l -> if gal q q' then p else cherche l in
      try cherche !t
      with Not_found ->
         t := (q,!n) :: !t;
         Queue.add q f;
         accept := final q :: !accept;
         incr n;
         !n - 1 in
   numro_of_tat dpart;
   while f <> Queue.create() do
      let d = Array.make (nombre_de_lettres ()) 0
      and q = Queue.take f in
      for i = 0 to nombre_de_lettres () - 1 do
         d.(i) <- numro_of_tat (delta q (lettre_of_numro i))
      done;
      delta' := d :: !delta'
   done;
   {delta = Array.of_list (List.rev !delta'); accept = Array.of_list(List.rev !accept)};;

let afd_of_af {initial = i; final = f} =
   construire_afd equal i af_delta (function q -> not (inter q f = vide));;

let afd_print {delta = d; accept = f} =
   print_string ".    |";
   for i = 0 to nombre_de_lettres() - 1 do
      print_string "  ";
      print_char(lettre_of_numro i)
   done;
   print_newline();
   print_string "______";
   for i = 1 to 3 * nombre_de_lettres() do print_char '_' done;
   for j = 0 to  Array.length d - 1 do
      print_newline();
      print_string(if f.(j) then ".F" else ". ");
      Printf.printf "%3d" j;
      print_char '|';
      for i = 0 to nombre_de_lettres() - 1 do
          Printf.printf "%3d" d.(j).(i)
      done
   done;
   print_newline();;

let af_print {initial = i; final = f} =
  let t_print s e =
    print_string (s ^ " : ");
    iter
      (
      function e -> Printf.printf "%i " e.numro
      )
      e;
    print_newline() in
  t_print "I" i;
  let tats_traits = ref (empty (fun n m -> n - m)) in
  let rec traite_tat {transitions = l; numro = n} =
    if not (mem n !tats_traits) then
      (
      tats_traits := add n !tats_traits;
      List.iter
        (
        function (c,e) ->
          Printf.printf "%i -%c-> %i\n" n c e.numro
        )
        l;
       List.iter
        (
        function (_,e) ->
          traite_tat e
        )
        l;
      ) in
  iter traite_tat i;
  t_print "F" f;;

let affiche_afd
    ?(ce = "DeepSkyBlue3") (* couleur tat*)
    ?(d = 7.)             (* diamtre tat *)
    ?(lf = 3.8)
    ?(eh = 16.)
    ?(ehp = 2.)
    ?(ehpp = 0.8)
    ?(lb = 5.)
    ?(mh = 0.1)
    ?(ct = "DeepSkyBlue4")
    ?(ra = 1.5)
    ?(ev = 5.5)
    ?(dv = 0.7)
    ?(hv = 3.)
    {delta = delta; accept = f} =
  let n = Array.length f in
  let g = Array.make n [] in
  for i = 0 to nombre_de_lettres() - 1 do
    let c = String.make 1 (lettre_of_numro i) in
    for q = 0 to n - 1 do
      let q' = delta.(q).(i) in
      try
        let s = List.assoc q' g.(q) in
        g.(q) <- (q',c ^ "," ^ s) :: List.remove_assoc q' g.(q)
      with Not_found -> g.(q) <- (q',c) :: g.(q)
    done
  done;
  let g' = Array.make n (0,[]) in
  for i = 0 to n - 1 do g'.(i) <- i,g.(i) done;
  let i = Array.make n false in
  i.(0) <- true;
  Graphe.draw_structuregraphique
     {
     Graphe.elementgraphique = g';
     Graphe.oriente = true;
     Graphe.flecherentrante = i;
     Graphe.flechesortante = f;
     Graphe.longueurfleche = lf;
     Graphe.fonctionsommet = (function s -> ce,string_of_int s);
     Graphe.fonctionarc = (function s -> ct,s);
     Graphe.ecarthorizontal = eh;
     Graphe.ecartvertical = ev;
     Graphe.diametresommet = d;
     Graphe.largeurboucle = lb;
     Graphe.distancevaluation = dv;
     Graphe.hauteurvaluation = hv;
     Graphe.trespetitecart = ehpp;
     Graphe.petitecart = ehp;
     Graphe.rayonangle = ra;
     Graphe.margehorizontale = mh;
     };;

let affiche_af
    ?(ce = "DeepSkyBlue3") (* couleur tat*)
    ?(d = 7.)             (* diamtre tat *)
    ?(lf = 3.8)
    ?(eh = 16.)
    ?(ehp = 2.)
    ?(ehpp = 0.8)
    ?(lb = 5.)
    ?(mh = 0.1)
    ?(ct = "DeepSkyBlue4")
    ?(ra = 1.5)
    ?(ev = 5.5)
    ?(dv = 0.7)
    ?(hv = 3.)
    {initial = i; final = f} =
  let nbtats = ref 0
  and t = ref []
  and table = Hashtbl.create 20
  and table' = Hashtbl.create 20 in
  let rec traite_tat {transitions = l; numro = n} =
    if not (Hashtbl.mem table' n) then
      (
      let n' = !nbtats in
      Hashtbl.add table n' n;
      Hashtbl.add table' n n';
      incr nbtats;
      List.iter
        (
        function (c,e) ->
          traite_tat e;
          t := (n',String.make 1 c,Hashtbl.find table' e.numro) :: !t
        )
        l
      ) in
  iter traite_tat i;
  let g = Array.make !nbtats [] in
  List.iter 
    (
    function q,c,q' -> 
      try
        let s = List.assoc q' g.(q) in
        g.(q) <- (q',c ^ "," ^ s) :: List.remove_assoc q' g.(q)
      with Not_found -> g.(q) <- (q',c) :: g.(q)
    ) 
    !t;
  let g' = Array.make !nbtats (0,[]) in
  for i = 0 to !nbtats - 1 do g'.(i) <- i,g.(i) done;
  let tb ens =
    let t = Array.make !nbtats false in
    iter (function e -> t.(Hashtbl.find table' e.numro) <- true) ens;
    t in
  Graphe.draw_structuregraphique
     {
     Graphe.elementgraphique = g';
     Graphe.oriente = true;
     Graphe.flecherentrante = tb i;
     Graphe.flechesortante = tb f;
     Graphe.longueurfleche = lf;
     Graphe.fonctionsommet = (function s -> ce,string_of_int s);
     Graphe.fonctionarc = (function s -> ct,s);
     Graphe.ecarthorizontal = eh;
     Graphe.ecartvertical = ev;
     Graphe.diametresommet = d;
     Graphe.largeurboucle = lb;
     Graphe.distancevaluation = dv;
     Graphe.hauteurvaluation = hv;
     Graphe.trespetitecart = ehpp;
     Graphe.petitecart = ehp;
     Graphe.rayonangle = ra;
     Graphe.margehorizontale = mh;
     };;

let afd_ntats {delta = d} = Array.length d;;

let am_of_afd ({delta = delta; accept = accept} as afd)=
   (* 'a afd -> 'a afd. Retourne l'afd minimal quivalent  afd. *)
   (* L'algorithme utilis est donn dans poly pages 143 et 144. *)
   let ne = afd_ntats afd
   and nl = nombre_de_lettres () in
   (* Marquage des paires d'tats non quivalents {i,j}, i < j. *)
   let t = Array.make_matrix (ne - 1) ne []
   and marqu = Array.make_matrix (ne - 1) ne false in
   let rec marquer_rec (i,j) =
      marqu.(i).(j) <- true;
      List.iter
         (function (i',j') -> if not marqu.(i').(j') then marquer_rec (i',j'))
         t.(i).(j) in
   for i = 0 to ne - 2 do for j = i + 1 to ne - 1 do
      if accept.(i) <> accept.(j) then marqu.(i).(j) <- true
   done done;
   for i = 0 to ne - 2 do for j = i + 1 to ne - 1 do
      let faire f =
      (* (int * int -> unit) -> unit.                     *)
      (* Applique f  tous les sommets adjacents  (i,j). *)
         for x = 0 to nl - 1 do
            let i' = delta.(i).(x)
            and j' = delta.(j).(x) in
            if i' < j' then f (i',j') else if i' > j' then f (j',i')
         done in
      if not marqu.(i).(j)
         then
            try
               faire
                  (function (i',j') -> if marqu.(i').(j') then raise Exit);
               faire
                  (function (i',j') ->
                      if t.(i').(j') = [] or List.hd t.(i').(j') <> (i,j)
                         then t.(i').(j') <- (i,j) :: t.(i').(j'))
            with Exit -> marquer_rec (i,j)
   done done;
   (* Construction de l'automate minimal  partir du tableau marqu. *)
   construire_afd
     (fun i j ->
         i = j or i < j & not marqu.(i).(j) or i > j & not marqu.(j).(i))
     0
     (fun q c -> delta.(q).(numro_of_lettre c))
     (function q -> accept.(q));;

(* ne figure pas dans poly *)

let rec set_of_list = function
   [] -> vide |
   e :: l -> add e (set_of_list l);;

let af_dfinir initial final =
   let i = set_of_list initial and f = set_of_list final in
   {initial = i; final = f; contient_mu = not (equal (inter i f) vide)};;

let af_of_afd ({delta = delta; accept = accept} as a) =
   let n = afd_ntats a in
   let t = Array.make n (nouvel_tat ()) in
   for q = 1 to n - 1 do t.(q) <- nouvel_tat () done;
   for q = 0 to n - 1 do for x = 0 to nombre_de_lettres () - 1 do
      ajoute_transition t.(q) (lettre_of_numro x) t.(delta.(q).(x))
   done done;
   let f = ref vide in
   for q = 0 to n - 1 do if accept.(q) then f := add t.(q) !f done;
   {initial = add t.(0) vide; final = !f; contient_mu = accept.(0)};;

let afd_complmentaire {accept = a} =
   for i = 0 to Array.length a - 1 do a.(i) <- not a.(i) done;;

let afd_produit prdicat {delta = d; accept = a} {delta = d'; accept = a'} =
   construire_afd
      (=)
      (0,0)
      (fun (q,q') c -> let i = numro_of_lettre c in d.(q).(i),d'.(q').(i))
      (function (q,q') -> prdicat a.(q) a'.(q'));;

let afd_union = afd_produit (fun b b' -> b or b')
and afd_inter = afd_produit (fun b b' -> b & b')
and afd_diff = afd_produit (fun b b' -> b & not b')
and afd_diffsym = afd_produit (fun b b' -> b & not b' or not b & b');;

let afd_vide {accept = a} =
   let rec b i = i = Array.length a or not a.(i) & b (i + 1) in
   b 0;;

let afd_inclus a a' = afd_vide (afd_diff a a')
and afd_gal a a' = afd_vide (afd_diffsym a a');;

let af_of_string s = af_of_er(er_of_string s);;
let afd_of_string s = afd_of_af(af_of_string s);;
let am_of_string s = am_of_afd(afd_of_string s);;
let afd_of_er s = afd_of_af(af_of_er s);;
let am_of_er s = am_of_afd(afd_of_er s);;
let am_of_af s = am_of_afd(afd_of_af s);;

(**************************************************************)
(* Sortie LaTeX de la reprsentation fonctionnelle d'un AFD   *)
(**************************************************************)

let fichier_courant = ref stdout;;

let open_file fichier = fichier_courant := open_out fichier
and close_file() = close_out !fichier_courant;;

let output_string s = (* entre open_file fichier et close_file() *)
  Pervasives.output_string !fichier_courant s;;

let output_int x = (* entre open_file fichier et close_file() *)
  output_string (string_of_int x);;

let output_afd position {delta = d; accept = a} =
   (* entre open_file fichier et close_file(); position est 't', 'b' ou 'c' *)
   let n = Array.length d
   and p = nombre_de_lettres() in
   output_string("$\\begin{array}[" ^ (String.make 1 position) ^ "]{c|");
   for i = 1 to n do output_string "c" done;
   output_string "}\n";
   output_string "&\\downarrow";
   if a.(0) then output_string "\\uparrow";
   for i = 1 to n - 1 do
      output_string "&";
      if a.(i) then output_string "\\uparrow"
   done;
   output_string "\\\\\n";
   for i = 1 to n do
      output_string "&";
      output_int i
   done;
   output_string"\\\\\\hline\n";
   for x = 0 to p - 1 do
      output_string (String.make 1(lettre_of_numro x));
      for i = 0 to n - 1 do
         output_string "&";
         output_int (d.(i).(x) + 1)
      done;
      if x < p - 1 then output_string "\\\\\n"
   done;
   output_string"\n\\end{array}$";;

let output_fichier_afd fichier afd =
   open_file fichier;
   output_afd 't' afd;
   close_file();;

