(************************** Dessins d'arbres *******************************)

type 'a arbre = {etiquette : 'a ; branches : 'a foret}
and 'a foret = 'a arbre list;;

type 'a arbrebinaire =
   Nil | Noeud of 'a arbrebinaire * 'a * 'a arbrebinaire;;

let rec calcule_franges l1 l2 a delta g1 d1 g2 d2 = 
  match (g1,d1,g2,d2) with
   (ag1 :: lg1),(ad1 :: ld1),(ag2 :: lg2),(ad2 :: ld2) ->
      let a', delta' =
         if a +. ag2 -. ad1 < 1.
            then 1. , delta +. 1. -. a -. ag2 +. ad1
            else a +. ag2 -. ad1 , delta in
      let lg1',ld2',delta'' =
        calcule_franges
          (ad1 +. l1 -. ag1) (ad2 +. l2 -. ag2) a' delta'  lg1 ld1 lg2 ld2 in
      ag1 :: lg1',ad2 :: ld2',delta'' |
   lg1, (ad1 :: ld1), [], [] -> lg1 , ad1 -. a -. l2 :: ld1 , delta |
   [], [], (ag2 :: lg2), ld2 -> ag2 +. l1 +. a :: lg2 , ld2 , delta |
   _ -> [] , [] , delta;;

let ecarts_arbre,           (* 'a arbre -> ('a * float) arbre *)
    ecarts_foret =          (* 'a foret -> ('a * float) foret *)
   let largeur f = List.fold_right (fun {etiquette=(_,d)} s -> d +. s) f 0. in
   let rec aa {etiquette = x;branches = f} =
        match f with
           [] -> {etiquette=(x,0.);branches=[]} , ([],[]) |
           _ -> let f',(lg,ld) = ff f in
                let e = largeur f' /. 2. in
                let rec transf e = function
                   [] -> [] |
                   {etiquette=(x,d);branches=f} :: l ->
                      {etiquette=(x,e);branches=f} :: (transf (e +. d) l) in
                {etiquette=(x,0.);branches=transf (-. e) f'} ,
                                                     ((-. e) :: lg , e :: ld)
   and ff = function
        [] -> failwith "ff" |
        [a] -> let a', c = aa a in
               [a'] , c |
        a :: f ->
           let {etiquette=(x,_);branches=f1},(lg1,ld1) = aa a
           and f',(lg2,ld2) = ff f in
           let lg,ld,delta =
              calcule_franges 0. (largeur f') 1. 1. lg1 ld1 lg2 ld2 in
           {etiquette=(x,delta);branches=f1} :: f' , (lg,ld) in
   (function a -> fst(aa a)) , (function f -> fst(ff f));;

let rec ecarts_totaux_arbreecarts {branches = f} =
   List.fold_right
      (
      fun ({etiquette = (_,e)} as a) (eg,ed) ->
         let eg',ed' = ecarts_totaux_arbreecarts a in
         min (e +. eg') eg , max (e +. ed') ed
      )
      f (0.,0.);;

let ecarts_totaux_foretecarts f =
   List.fold_right
      (
      fun ({etiquette = (_,e)} as a) (eg,ed) ->
         let eg',ed' = ecarts_totaux_arbreecarts a in
         min eg' (e +. eg) , max ed' (e +. ed)
      )
      f (0.,0.);;

let rec hauteur_arbre {branches = f} = hauteur_foret f + 1
and hauteur_foret f = List.fold_right (function a -> max (hauteur_arbre a)) f (-1)
and largeur_arbre a =
   let eg,ed = ecarts_totaux_arbreecarts (ecarts_arbre a) in ed -. eg
and largeur_foret f = let eg,ed =
   ecarts_totaux_foretecarts (ecarts_foret f) in ed -. eg;;

let dimensions_arbre a = largeur_arbre a,float_of_int(hauteur_arbre a)
and dimensions_foret f = largeur_foret f,float_of_int(hauteur_foret f);;

let ecarts_arbrebinaire a = (*'a arbrebinaire -> ('a * float) arbrebinaire*)
   let rec ab = function
      Nil -> Nil , ([],[]) |
      Noeud(Nil,x,Nil) -> Noeud(Nil,(x,0.),Nil) , ([],[]) |
      Noeud(u,x,Nil) ->
         let u', (lg,ld) = ab u in
         Noeud(u',(x,0.5),Nil) , (-.0.5 :: lg,-.0.5 :: ld) |
      Noeud(Nil,x,v) ->
         let v', (lg,ld) = ab v in
         Noeud(Nil,(x,0.5),v') , (0.5 :: lg,0.5 :: ld) |
      Noeud(u,x,v) ->
         let u' , (lg1,ld1) = ab u
         and v' , (lg2,ld2) = ab v in
         let lg,ld,d = calcule_franges 0. 0. 1. 1. lg1 ld1 lg2 ld2 in
         Noeud(u',(x,d /. 2.),v'),
         (-. d /. 2. :: lg,d /. 2. :: ld) in
   fst(ab a);;

let rec ecarts_totaux_arbrebinaireecarts = function
   Nil -> 0.,0. |
   Noeud(u,(_,e),v) -> let eg,ed = ecarts_totaux_arbrebinaireecarts u
                       and eg',ed' = ecarts_totaux_arbrebinaireecarts v in
                       match u,v with
                          Nil,Nil -> 0.,0. |
                          Nil,_ -> min 0. (eg' +. e),ed' +. e |
                          _,Nil -> eg -. e,max 0. (ed -. e) |
                          _ -> min (eg -. e) (eg' +. e),
                               max (ed -. e) (ed' +. e);;
let rec hauteur_arbrebinaire = function
   Nil -> -1 |
   Noeud(u,_,v) -> max (hauteur_arbrebinaire u + 1) (hauteur_arbrebinaire v + 1);;

let largeur_arbrebinaire a =
   let eg,ed = ecarts_totaux_arbrebinaireecarts (ecarts_arbrebinaire a) in
   ed -. eg;;

let dimensions_arbrebinaire a =
   largeur_arbrebinaire a,float_of_int(hauteur_arbrebinaire a);;

type 'a elementgraphique =
   Arbre of 'a arbre |
   Foret of 'a foret |
   Arbrebinaire of 'a arbrebinaire;;

type 'a structuregraphique =
   {
   elementgraphique    : 'a elementgraphique;
   fonctionnoeud       : 'a -> float -> float -> unit;
   fonctionarc         : 'a -> 'a -> float -> float -> float -> float -> unit;
   ecarthorizontal     : float;
   ecartvertical       : float;
   largeurnoeud        : float;
   hauteurnoeud        : float
   };;

let dimensions_structuregraphique
   {
   elementgraphique = eg;
   ecarthorizontal  = eh;
   ecartvertical    = ev;
   largeurnoeud     = ln;
   hauteurnoeud     = hn
   } =
   let l,h = match eg with
      Arbre a -> dimensions_arbre a |
      Foret f -> dimensions_foret f |
      Arbrebinaire ab -> dimensions_arbrebinaire ab in
   l *. eh +. ln,h *. ev +. hn;;

let draw_structuregraphique
   {
   elementgraphique = eg;
   fonctionnoeud    = fn;
   fonctionarc      = fa;
   ecarthorizontal  = eh;
   ecartvertical    = ev;
   largeurnoeud     = ln;
   hauteurnoeud     = hn
   } x y =
   let rec da {etiquette=(e,_);branches=f} x y =
       List.iter
          (
          function {etiquette=(e',ecart)} as a ->
             let x' = x +. ecart *. eh
             and y' = y +. ev in
             fa e e' x y x' y';
             da a x' y'
          )
          f;
       fn e x y;
       () in
   match eg with
      Arbre a ->
         let a' = ecarts_arbre a in
         da a' (x -. eh *. fst(ecarts_totaux_arbreecarts a') +. ln /.2.)
                                                             (y +. hn /. 2.) |
      Foret f ->
         let f' = ecarts_foret f in
         List.fold_left
            (
            fun x ({etiquette=(_,ecart)} as a) ->
               da a x (y +. hn /. 2.);
               x +. ecart *. eh
            )
            (x -. eh *. fst(ecarts_totaux_foretecarts f') +. ln /. 2.)
            f';
         () |
      Arbrebinaire ab ->
         let rec dab x y  = function
            Nil -> () |
            Noeud(u,(etiq,ecart),v) ->
                  let f e = function
                     Nil -> () |
                     Noeud(_,(etiq',_),_) as w ->
                           let x' = x +. e *. eh
                           and y' = y +. ev in
                           fa etiq etiq' x y x' y';
                           dab x' y' w in
                  f (-.ecart) u;
                  f ecart v;
                  fn etiq x y in
         let ab' = ecarts_arbrebinaire ab in
         dab (x -. eh *. fst(ecarts_totaux_arbrebinaireecarts ab') +. ln /. 2.)
                                                          (y +. hn /. 2.) ab';;

let rec arbrebinaire_of_arbre {etiquette = e; branches = f} =
   match f with
      [] -> Nil |
      [a;b] -> Noeud(arbrebinaire_of_arbre a,e,arbrebinaire_of_arbre b) |
      _ -> failwith "arbrebinaire_of_arbre";;

let rec arbre_of_arbrebinaire x =  function
   Nil -> {etiquette = x;branches = []} |
   Noeud(u,e,v) ->
      {etiquette = e;
       branches = [arbre_of_arbrebinaire x u;arbre_of_arbrebinaire x v]};;

let arbre_of_string,foret_of_string =
   let couple s =
      let s = s ^ "#"
      and k = ref 0 in
      let nth() = s.[!k] in
      let rec arbre() =
         let l = ref [] in
         while nth() <> '(' do
            l := nth() :: !l;
            incr k
         done;
         let a = String.create (List.length !l) in
         List.fold_left (fun i c -> a.[i] <- c; i + 1) 0 (List.rev !l);
         {etiquette = a;branches = foret()}
      and foret() =
         incr k;
         let l = ref [] in
         while nth() <> ')' do
            l := arbre() :: !l
         done;
         incr k;
         List.rev !l in
      arbre,foret in
   (function s -> fst (couple s) ()),(function s -> snd (couple s) ());;

let arbrebinaire_of_string s = arbrebinaire_of_arbre(arbre_of_string s);;

let affiche_elementgraphique 
    ?(ca = "DeepSkyBlue4")
    ?(cn = "DeepSkyBlue3")
    ?(d  = 4.)
    ?(eh = 8.)
    ?(ev = 6.) f eg =
  let sg =
    {
    elementgraphique = eg;
    fonctionnoeud =
      (
      fun n x y ->
        Es.draw_circle ~fill:cn ~text:(f n) ~d:d x y
      );
    fonctionarc =
      (
      fun _ _ x1 y1 x2 y2 -> Es.draw_line ~fill:ca x1 y1 x2 y2
      );
    ecarthorizontal = eh;
    ecartvertical = ev;
    largeurnoeud = 1.5 *. d;
    hauteurnoeud = 1.5 *. d
    } in
  let l,h = dimensions_structuregraphique sg in
  Es.open_canvas l h;
  draw_structuregraphique sg 0. 0.;
  Es.close_canvas();;









